%Blakely-Simpson (1986, Geophysics) algorithm
% reading of the GS Surfer ASCII grid into the matrix field
clear all; close all; clc;
rowsm = 0; columnsn = 0; minx = 0; maxx= 0; miny = 0; maxy= 0; minf = 0; maxf= 0;
% opening the file
[fname,pname] = uigetfile('*.grd','input of data (Surfer ASCII grid)');
fid = fopen(fname,'r');
% reading of the GS ASCII grid header (rows, columns, minx, maxx, miny, maxy, minf, maxf)
tline = fgetl(fid); row1 = tline;
tline = fgetl(fid); row2 = tline; [token,rem] = strtok(row2);
rowsm = str2double(token); columnsn = str2double(rem);
tline = fgetl(fid); row3 = tline; [token,rem] = strtok(row3);
minx = str2double(token); maxx = str2double(rem);
tline = fgetl(fid); row4 = tline; [token,rem] = strtok(row4);
miny = str2double(token); maxy = str2double(rem);
tline = fgetl(fid); row5 = tline; [token,rem] = strtok(row5);
minf = str2double(token); maxf = str2double(rem);
% reading of the main field of the grid - into the matrix A
[A,count] = fscanf(fid,'%f',[rowsm,columnsn]);
% closing the file
status = fclose(fid);
field = A';
% size of the readed matrix: columnsn x rowsm
[columnsn,rowsm] = size(A);

%generation of x- and y-coordinates (one-row matrices xcor and ycor)
xcor=(minx:(maxx-minx)/(columnsn-1):maxx);
ycor=(miny:(maxy-miny)/(rowsm-1):maxy);
dx = xcor(2) - xcor(1); dy = ycor(2) - ycor(1);

% plotting filled contour map of the imported field
contourf(xcor,ycor,field); 
%adding the colorbar
colorbar;
%labeling the axes
xlabel('x [m]','FontSize',10); ylabel('y [m]','FontSize',10); legend('CBA 2.67');

%evaluation of x- and y-derivative in the space domain
for i=1:rowsm-1  %begin of the cycle for changing rows
    for j=1:columnsn-1  %begin of the cycle for changing columns 
        hg_x(i,j) = (field(i,j+1)-field(i,j))/dx;
        hg_y(i,j) = (field(i+1,j)-field(i,j))/dy;
    end; %end of the cycle for j
end; %end of the cycle for i
% managing the same matrices dimensions (identical with the original)
hg_x(:,columnsn) = hg_x(:,columnsn-1); hg_y(:,columnsn) = hg_y(:,columnsn-1); 
hg_x(rowsm,:) = hg_x(rowsm-1,:); hg_y(rowsm,:) = hg_y(rowsm-1,:);

%evaluation of the horizontal gradient
hg = sqrt(hg_x.^2 + hg_y.^2);

% plotting filled contour maps of the calculated x- and y-derivatives
figure, contourf(xcor,ycor,hg_x); colorbar; xlabel('x [m]','FontSize',10); ylabel('y [m]','FontSize',10); legend('hg_x');
figure, contourf(xcor,ycor,hg_y); colorbar; xlabel('x [m]','FontSize',10); ylabel('y [m]','FontSize',10); legend('hg_y');
figure, contourf(xcor,ycor,hg); colorbar; xlabel('x [m]','FontSize',10); ylabel('y [m]','FontSize',10); legend('hg');

% Blakely - Simpson method
solution_number = 1;
for i=2:rowsm-1  %begin of the cycle for changing rows
    for j=2:columnsn-1  %begin of the cycle for changing columns
        %starting value of presence (containing how many conditions are fulfilled)
        presence(solution_number) = 0; xcor_solution(solution_number) = xcor(j); ycor_solution(solution_number) = ycor(i);
        %1. condition - in N-S direction
        if hg(i-1,j)<hg(i,j) & hg(i,j)>hg(i+1,j)
           presence(solution_number) = presence(solution_number) + 1;
        end;
        %2. condition - in NW-SE direction
        if hg(i-1,j-1)<hg(i,j) & hg(i,j)>hg(i+1,j+1)
           presence(solution_number) = presence(solution_number) + 1;
        end;
        %3. condition - in W-E direction
        if hg(i,j-1)<hg(i,j) & hg(i,j)>hg(i,j+1)
           presence(solution_number) = presence(solution_number) + 1;
        end;
        %4. condition - in SW-NE direction
        if hg(i+1,j-1)<hg(i,j) & hg(i,j)>hg(i-1,j+1)
           presence(solution_number) = presence(solution_number) + 1;
        end;
        %increasing the solution number
        solution_number = solution_number + 1;
    end; %end of the cycle for j
end; %end of the cycle for i
%correcting the solution number
solution_number = solution_number - 1;

%displaying the final number of solutions
disp('final number of solutions'); disp(solution_number);

%solutions selection
selected_solution_number = 1;
for k = 1:solution_number
    if presence(k)>2  % level of solution acceptation
        selected_presence(selected_solution_number) = presence(k);
        xcor_selected_solution(selected_solution_number) = xcor_solution(k);
        ycor_selected_solution(selected_solution_number) = ycor_solution(k);
        %increasing the selected solution number
        selected_solution_number = selected_solution_number + 1;
    end;
end;
%correcting the selected solution number
selected_solution_number = selected_solution_number - 1;

%displaying the final number of solutions
disp('final selected number of solutions'); disp(selected_solution_number);

%plotting selected solutions (filled squares with no connecting lines)
plot(xcor_selected_solution,ycor_selected_solution,' s', 'MarkerEdgeColor','k', 'MarkerFaceColor','k','MarkerSize', 4);
xlabel('x [m]','FontSize',10); ylabel('y [m]','FontSize',10); legend('B-S solutions');

% writing selected solutions into a file
% creating a matrix from solution vectors
solXYP = [xcor_selected_solution' ycor_selected_solution' selected_presence'];
save  Blakely-Simpson_solutions.dat  solXYP  -ascii  -double;

%end of the script