% Last and Kubik (1983) algorithm - testing on models
clear all; close all; clc;

gamma=6.67E-11;

% reading of the model from SRF ASCII grid (densities)
[densities, xmin, xmax, ymin, ymax]= grid_read_SRF_ASCII('modelrenault1.grd'); %example from page 716 (Fig. 2)
[rows,cols] = size(densities);
% N - number of line points
% M - number of model elements
N = cols; M = rows.*cols; dx=(xmax-xmin)/(N); dy=(ymax-ymin)/(rows-1); %nerozumiem tejto rows-1...?
h = dy; d = dx; %parameters of the model elements
% generation of the x- and z-coordinates of the line points on the model surface
for i = 1:cols
    xi(i)= xmin+dx/2+(i-1)*dx; zi(i) = 0;
end;
% generation of the x- and z-coordinates of the line points on the model surface
jj=1;
for j = 1:rows
    for i = 1:cols
    xj(jj)= xmin+dx/2+(i-1)*dx;
    zj(jj)= ymin+dy/2+(j-1)*dy; 
    nyjh(jj)=densities(j,i); jj = jj + 1;
    end;
end;
nyj = nyjh';

%displaying the test model (densities in xz-section)
 figure, contourf(flipud(densities));legend('tested model');
 
% generation of the a-matrix (using effect of 2D rectangular prism)
 for i=1:N
  for j=1:M
  r1 = sqrt( (zi(i)+zj(j)-h/2)*(zi(i)+zj(j)-h/2) + (xi(i)-xj(j)+d/2)*(xi(i)-xj(j)+d/2) ); 
  r2 = sqrt( (zi(i)+zj(j)+h/2)*(zi(i)+zj(j)+h/2) + (xi(i)-xj(j)+d/2)*(xi(i)-xj(j)+d/2) );
  r3 = sqrt( (zi(i)+zj(j)-h/2)*(zi(i)+zj(j)-h/2) + (xi(i)-xj(j)-d/2)*(xi(i)-xj(j)-d/2) );
  r4 = sqrt( (zi(i)+zj(j)+h/2)*(zi(i)+zj(j)+h/2) + (xi(i)-xj(j)-d/2)*(xi(i)-xj(j)-d/2) );
  theta1 = atan( (xi(i)-xj(j)+d/2)/(zi(i)+zj(j)-h/2) );
  theta2 = atan( (xi(i)-xj(j)+d/2)/(zi(i)+zj(j)+h/2) );
  theta3 = atan( (xi(i)-xj(j)-d/2)/(zi(i)+zj(j)-h/2) );
  theta4 = atan( (xi(i)-xj(j)-d/2)/(zi(i)+zj(j)+h/2) );
  a(i,j) = 2*gamma*( (xi(i)-xj(j)+d/2)*log((r2*r3)/(r1*r4))+d*log(r4/r3)-...
                     (zj(j)+h/2)*(theta4-theta2)+(zj(j)-h/2)*(theta3-theta1) );
  end; %(for j=1:M)
 end; %(for i=1:N) 

 % direct problem (modelled field is in the g-matrix)
g = a*nyj;

% **************** INVERSION *******************
% simpliest LSQ-solution (not working well, page 716 (Fig. 2b)
lsqdens = a'*inv(a*a')*g;
 jj=1;
   for j = 1:rows
      for i = 1:cols
      visu0(j,i)=lsqdens(jj); jj = jj + 1;
      end
   end;
 figure, contourf(flipud(visu0));legend('simplest LSQ inversion');

% simple Tikhonov regularization, little bit better, but still very wrong (shallow)
lam = 0.1; %regularization parameter, usually a positive number between 1E-10 and 1E+10
[mw,nw]=size(a*a'); Id = eye(mw,nw);% genering the identity matrix Id
tichregdens = a'*inv(a*a'+lam*Id)*g;
 jj=1;
   for j = 1:rows
      for i = 1:cols
      visu0(j,i)=tichregdens(jj); jj = jj + 1;
      end
   end;
 figure, contourf(flipud(visu0));legend('Tikhonov inversion without compacting');
 
%Last-Kubik 1983 algorithm (simpliest version)
% preparation:
l_0 = 0.01; % estimated noise/signal ratio 
eps = 1E-1; % better results than with 1E-11, recommended in page 715, right column, central part
w_ny = eye(M,M);   w_ny_inv = inv(w_ny); % genering the identity matrix Id as the first iteration for the weighting matrix w_ny
w_e_inv = l_0*l_0*a*w_ny_inv*a'; % weighting (noise e) matrix in Eq.11 an Eq.17

% number of iterations
number_iterations = 15;
for k=1:number_iterations % iteration cycle
   disp('interation nr.:'); k    
   LKdens = w_ny_inv*a'*inv(a*w_ny_inv*a'+w_e_inv)*g; %main solution, Eq. 15
   LKdensT = LKdens';
   % results visualisation after each iteration step
   jj=1;
   for j = 1:rows
      for i = 1:cols
      visu1(j,i)=LKdensT(jj); jj = jj + 1;
      end
   end;
   infoonmethod='Last and Kubik, it.Nr.: '; fordisplay=append(infoonmethod, num2str(k));
   figure, contourf(flipud(visu1)); legend(fordisplay);
   % adapting the weighting matrix w_ny, Eq.16
   for j=1:M
       w_ny_inv(j,j) = LKdens(j)*LKdens(j)+eps; %empiricky overene, ze takto vyratana inverzna je identicka s tou vyratanou podla rovnice (8) a potom invertovanou
   end; %(for j=1:M)
      w_e_inv = l_0*l_0*a*w_ny_inv*a'; % adapting the weighting matrix w_ne, Eq.11
end; %(for k=1:10)

% exporting results into an ASCII SRF grid
grid_write_SRF_ASCII(visu1,xmin,xmax,ymin,ymax,'LKdens.grd');

% writing of determined solutions into a file
% creating a matrix from solution vectors
solXZ = [xi' g];
save  modeled_field.dat  solXZ  -ascii  -double;
% END
